import { HttpService, Injectable } from '@nestjs/common';
import { Logger } from '@nestjs/common';
import { Observable } from 'rxjs';
import { AxiosResponse } from 'axios';
import { Cron, CronExpression } from '@nestjs/schedule';
import * as _ from 'lodash';
import * as https from 'https';
import { InjectModel } from '@nestjs/mongoose';
import { Model } from 'mongoose';
import { Structure, StructureDocument } from '../schemas/structure.schema';
import { ApticStructure } from '../schemas/aptic-structure.schema';
import { Address } from '../schemas/address.schema';
import { UsersService } from '../../users/users.service';
import { ApticCatalog } from '../schemas/aptic-catalog.schema';
import { CategoriesFormationsService } from '../../categories/services/categories-formations.service';
import { StructuresSearchService } from './structures-search.service';

@Injectable()
export class ApticStructuresService {
  constructor(
    private readonly httpService: HttpService,
    private readonly userService: UsersService,
    private readonly categoriesFormationsService: CategoriesFormationsService,
    private structuresSearchService: StructuresSearchService,
    @InjectModel(Structure.name) private structureModel: Model<StructureDocument>
  ) {}

  /**
   * Get all aptic structures in the given postal codes.
   * @param postalCodeData
   */
  public formatApticStructures(postalCodeData: any[]): void {
    // Get all postal code in one array
    const postalCodeArray = _.flatten(postalCodeData.map((data) => data.codesPostaux));

    // Call APTIC Api's
    postalCodeArray.map((postalCode) => {
      return this.getApticStructures(postalCode).subscribe(
        (res) => {
          res.data.presencePoints.forEach((structure) => {
            // Call aptic api for offer
            this.getApticStructureOffer(structure.catalogs[0]).subscribe(
              (serviceData) => {
                structure.services = serviceData.data.services;
                // Create structure
                this.createApticStructures(structure);
              },
              (err) => {
                Logger.log(err);
              }
            );
          });
        },
        (err) => {
          Logger.log(`getApticStructures error on postal code: ${postalCode}. Code: ${err}`);
        }
      );
    });
  }

  /**
   * Create a structure for app database given an aptic structure
   * @param structure ApticStructure
   */
  private async createApticStructures(structure: ApticStructure): Promise<any> {
    this.structureAlreadyExist(structure).then(async (exist) => {
      if (!exist) {
        Logger.log(`Create structure : ${structure.name}`, 'ApticStructuresService - createApticStructures');
        const createdStructure = new this.structureModel();
        // Known fields
        createdStructure.structureName = structure.name;
        createdStructure.contactPhone = structure.phone;
        // Unkown fields (but mandatory)
        createdStructure.contactMail = 'unknown@unknown.com';
        createdStructure.labelsQualifications = ['passNumerique'];
        createdStructure.structureType = 'autre';
        createdStructure.pmrAccess = false;
        createdStructure.accessModality = ['accesLibre'];
        createdStructure.publics = ['toutPublic'];
        createdStructure.accountVerified = true;
        createdStructure.freeWorkShop = false;
        createdStructure.nbComputers = null;
        createdStructure.nbPrinters = null;
        createdStructure.nbScanners = null;
        createdStructure.nbTablets = null;
        createdStructure.nbNumericTerminal = null;
        // Address
        createdStructure.coord = [structure.address.gpsLng, structure.address.gpsLat];
        createdStructure.address = this.formatAddress(structure);
        // Set structure offer
        createdStructure.parentingHelp = await this.setModules(structure, 'parentingHelp');
        createdStructure.baseSkills = await this.setModules(structure, 'baseSkills');
        createdStructure.accessRight = await this.setModules(structure, 'accessRight');
        createdStructure.socialAndProfessional = await this.setModules(structure, 'socialAndProfessional');
        createdStructure.digitalCultureSecurity = await this.setModules(structure, 'digitalCultureSecurity');
        createdStructure.save();
        this.structuresSearchService.indexStructure(createdStructure);
        // Send admin weird structure mail
        this.verifyDuplication(createdStructure);
      }
    });
  }

  /**
   * Given an aptic structure, this method return the corresponding services id by category.
   * @param structure ApticStructure
   * @param moduleCategory string
   * @returns Promise<string[]>
   */
  private async setModules(structure: ApticStructure, moduleCategory: string): Promise<string[]> {
    const referentialIds = await this.categoriesFormationsService
      .findOne(moduleCategory)
      .then((referential) => referential.modules.map((element) => element.id));

    return structure.services
      .filter((service) => referentialIds.includes(service.code))
      .map((service) => {
        return service.code;
      });
  }

  /**
   * Verifiy if an aptic structure already exist in database.
   *  - If it's true return true and update fields if needed
   *  - If it's false return false
   * @param structure ApticStructure
   * @returns boolean
   */
  private async structureAlreadyExist(structure: ApticStructure): Promise<boolean> {
    let existingStructure = await this.structureModel
      .findOne({
        structureName: { $regex: structure.name, $options: 'i' },
      })
      .exec();
    // Check without regex for case like 'TINEBRA*DANIEL/DANIEL/'
    if (!existingStructure) {
      existingStructure = await this.structureModel.findOne({ structureName: structure.name }).exec();
    }

    if (existingStructure) {
      // Add aptic label if it's not the case
      if (!existingStructure.labelsQualifications.includes('passNumerique')) {
        existingStructure.labelsQualifications.push('passNumerique');
      }
      // Update service offer
      existingStructure.parentingHelp = _.unionWith(
        existingStructure.parentingHelp,
        await this.setModules(structure, 'parentingHelp'),
        _.isEqual
      );
      existingStructure.baseSkills = _.unionWith(
        existingStructure.baseSkills,
        await this.setModules(structure, 'baseSkills'),
        _.isEqual
      );
      existingStructure.accessRight = _.unionWith(
        existingStructure.accessRight,
        await this.setModules(structure, 'accessRight'),
        _.isEqual
      );
      existingStructure.socialAndProfessional = _.unionWith(
        existingStructure.socialAndProfessional,
        await this.setModules(structure, 'socialAndProfessional'),
        _.isEqual
      );
      existingStructure.digitalCultureSecurity = _.unionWith(
        existingStructure.digitalCultureSecurity,
        await this.setModules(structure, 'digitalCultureSecurity'),
        _.isEqual
      );
      existingStructure.save();
      return true;
    }
    return false;
  }

  /**
   * Get Metropole new aptic structure evey week. For testing, please change the expression
   */
  @Cron(CronExpression.EVERY_WEEK)
  public getMetopoleMunicipality(): void {
    const req =
      'https://download.data.grandlyon.com/ws/grandlyon/adr_voie_lieu.adrcomgl/all.json?maxfeatures=-1&start=1';
    Logger.log(`Request : ${req}`, 'ApticStructuresService - getMetopoleMunicipality');
    this.httpService.get(encodeURI(req)).subscribe(
      (data) => {
        const inseeArray = data.data.values.map((municipality) => {
          return this.getPostalCodeWithINSEE(municipality.insee).toPromise();
        });
        Promise.all(inseeArray).then((inseData) => {
          const postalCodeArray = inseData.map((cpData: { data; config }) => {
            return cpData.data;
          });
          this.formatApticStructures(postalCodeArray);
        });
      },
      (err) => Logger.error(err)
    );
  }

  public getPostalCodeWithINSEE(inseeCode: string): Observable<AxiosResponse<any>> {
    const req = `https://geo.api.gouv.fr/communes/${inseeCode}?fields=codesPostaux&format=json`;
    Logger.log(`Request : ${req}`, 'ApticStructuresService - getMetopoleMunicipality');
    return this.httpService.get(encodeURI(req));
  }

  public getApticStructures(postalCodeData: string): Observable<AxiosResponse<{ presencePoints: ApticStructure[] }>> {
    const req = `https://aptisearch-api.aptic.fr/v1/postal-code/${postalCodeData}`;
    Logger.log(`Request : ${req}`, 'ApticStructuresService');
    return this.httpService.get(req, {
      headers: {
        api_key: process.env.APTIC_TOKEN,
      },
    });
  }

  public getApticStructureOffer(catalogId: string): Observable<AxiosResponse<{ services: ApticCatalog[] }>> {
    const agent = new https.Agent({
      rejectUnauthorized: false,
    });
    const req = `https://aptisearch-api.aptic.fr/v1/catalog/${catalogId}/services`;
    Logger.log(`Request : ${req}`, 'ApticStructuresService');
    return this.httpService.get(req, {
      httpsAgent: agent,
      headers: {
        api_key: process.env.APTIC_TOKEN,
      },
    });
  }

  private async verifyDuplication(createdStructure: StructureDocument): Promise<void> {
    const sameAddrStructure = await this.structureModel
      .findOne({
        _id: { $ne: createdStructure._id },
        address: createdStructure.address,
      })
      .exec();
    if (sameAddrStructure) {
      this.userService.sendAdminApticStructureMail(createdStructure.structureName, sameAddrStructure.structureName);
    }
  }

  /**
   * Format aptic structure address
   */
  private formatAddress(structure: ApticStructure): Address {
    const address = new Address();
    const regexWithSpace = /\d+\s/g; // NOSONAR
    const regex = /\d+/g; // NOSONAR
    if (structure.address.main.match(regex)) {
      address.numero = structure.address.main.match(regex)[0];
      address.street = structure.address.main.replace(regexWithSpace, '');
    } else {
      address.street = structure.address.main;
    }
    address.commune = structure.address.city;
    return address;
  }
}
