import { Injectable } from '@nestjs/common';
import { ElasticsearchService } from '@nestjs/elasticsearch';
import { structureDto } from '../dto/structure.dto';
import { StructureDocument } from '../schemas/structure.schema';
import { StructureSearchBody } from '../interfaces/structure-search-body.interface';
import { StructureSearchResult } from '../interfaces/structure-search-response.interface';

@Injectable()
export class StructuresSearchService {
  private index = 'structures';

  constructor(private readonly elasticsearchService: ElasticsearchService) {}

  public async indexStructure(structure: StructureDocument): Promise<StructureDocument> {
    this.elasticsearchService.index<StructureSearchResult, StructureSearchBody>({
      index: this.index,
      id: structure._id,
      body: {
        structureName: structure.structureName,
        structureType: structure.structureType,
        structureId: structure._id,
        address: structure.address,
        description: structure.description,
      },
    });
    return structure;
  }

  public async createStructureIndex(): Promise<any> {
    return this.elasticsearchService.indices.create({
      index: this.index,
    });
  }

  public async dropIndex(): Promise<any> {
    if (
      (
        await this.elasticsearchService.indices.exists({
          index: this.index,
        })
      ).body
    ) {
      return this.elasticsearchService.indices.delete({
        index: this.index,
      });
    }
  }

  public async deleteIndexStructure(structure: StructureDocument): Promise<StructureDocument> {
    this.elasticsearchService.delete<StructureSearchResult, StructureSearchBody>({
      index: this.index,
      id: structure._id,
    });
    return structure;
  }

  public async search(searchString: string): Promise<StructureSearchBody[]> {
    searchString = searchString ? searchString + '*' : '*';
    const { body } = await this.elasticsearchService.search<StructureSearchResult>({
      index: this.index,
      body: {
        from: 0,
        size: 200,
        query: {
          query_string: {
            analyze_wildcard: 'true',
            query: searchString,
            fields: ['structureName^5', 'structureType^5', 'address.commune^10', 'description'],
            fuzziness: 'AUTO',
          },
        },
      },
    });
    const maxScore = Math.max.apply(
      Math,
      body.hits.hits.map(function (hit) {
        return hit._score;
      })
    );
    const sortedHits = body.hits.hits.filter(function (elem) {
      return elem._score >= maxScore / 1.5;
    });
    return sortedHits.map((item) => item._source);
  }

  public async update(structure: structureDto, id: string): Promise<any> {
    return this.elasticsearchService.update({
      index: this.index,
      id: id,
      body: {
        doc: {
          structureName: structure.structureName,
          structureType: structure.structureType,
          structureId: id,
          address: structure.address,
          description: structure.description,
        },
      },
    });
  }
}
