import { HttpException, HttpService, Injectable, HttpStatus, Logger } from '@nestjs/common';
import { InjectModel } from '@nestjs/mongoose';
import { Types, Model, FilterQuery, DocumentDefinition } from 'mongoose';
import { Observable } from 'rxjs';
import { AxiosResponse } from 'axios';
import { Structure, StructureDocument } from '../schemas/structure.schema';
import * as ejs from 'ejs';
import { structureDto } from '../dto/structure.dto';
import { UsersService } from '../../users/users.service';
import { User } from '../../users/schemas/user.schema';
import { MailerService } from '../../mailer/mailer.service';
import { Cron, CronExpression } from '@nestjs/schedule';
import { DateTime } from 'luxon';
import { IUser } from '../../users/interfaces/user.interface';
import * as _ from 'lodash';
import { OwnerDto } from '../../users/dto/owner.dto';
import { StructuresSearchService } from './structures-search.service';
import { CategoriesFormationsModule } from '../../categories/schemas/categoriesFormationsModule.schema';
import { CategoriesModule } from '../../categories/categories.module';
import { CategoriesAccompagnement } from '../../categories/schemas/categoriesAccompagnement.schema';
import { CategoriesFormations } from '../../categories/schemas/categoriesFormations.schema';
import { CategoriesOthers } from '../../categories/schemas/categoriesOthers.schema';

@Injectable()
export class StructuresService {
  constructor(
    private readonly httpService: HttpService,
    private readonly userService: UsersService,
    private readonly mailerService: MailerService,
    private structuresSearchService: StructuresSearchService,
    @InjectModel(Structure.name) private structureModel: Model<StructureDocument>
  ) {}

  async initiateStructureIndex(): Promise<StructureDocument[]> {
    await this.structuresSearchService.dropIndex();
    await this.structuresSearchService.createStructureIndex();
    return this.populateES();
  }

  async searchForStructures(text: string, filters?: Array<any>): Promise<StructureDocument[]> {
    const results = await this.structuresSearchService.search(text);
    const ids = results.map((result) => result.structureId);
    if (!ids.length) {
      return [];
    }
    //we match ids from Elasticsearch with ids from mongoDB (and filters) and sort the result according to ElasticSearch order.
    if (filters.length > 0) {
      return (
        await this.structureModel
          .find({
            _id: { $in: ids },
            $and: [...this.parseFilter(filters), { deletedAt: { $exists: false }, accountVerified: true }],
          })
          .exec()
      ).sort((a, b) => ids.indexOf(a.id) - ids.indexOf(b.id));
    } else {
      return (
        await this.structureModel
          .find({
            _id: { $in: ids },
            $and: [{ deletedAt: { $exists: false }, accountVerified: true }],
          })
          .exec()
      ).sort((a, b) => ids.indexOf(a.id) - ids.indexOf(b.id));
    }
  }

  public async create(idUser: string, structure: structureDto): Promise<Structure> {
    const user = await this.userService.findOne(idUser);
    if (!user) {
      throw new HttpException('Invalid profile', HttpStatus.NOT_FOUND);
    }
    const createdStructure = new this.structureModel(structure);
    createdStructure._id = Types.ObjectId();
    await createdStructure.save();
    await this.getStructurePosition(createdStructure).then(async (postition: StructureDocument) => {
      return this.structuresSearchService.indexStructure(
        await this.structureModel
          .findByIdAndUpdate(Types.ObjectId(createdStructure._id), {
            address: postition.address,
            coord: postition.coord,
          })
          .exec()
      );
    });
    user.structuresLink.push(createdStructure._id);
    user.save();

    // Senc admin notification mail
    this.sendAdminStructureNotification(
      createdStructure,
      this.mailerService.config.templates.adminStructureCreate.ejs,
      this.mailerService.config.templates.adminStructureCreate.json
    );

    return createdStructure;
  }

  public async search(searchString: string, filters?: Array<any>): Promise<Structure[]> {
    if (searchString && filters) {
      return this.structureModel
        .find({
          $and: [
            ...this.parseFilter(filters),
            { $text: { $search: searchString }, deletedAt: { $exists: false }, accountVerified: true },
          ],
        })
        .exec();
    } else if (filters) {
      return this.structureModel
        .find({ $and: [{ $and: this.parseFilter(filters), deletedAt: { $exists: false }, accountVerified: true }] })
        .exec();
    } else {
      return this.structureModel
        .find({
          $and: [{ $or: [{ $text: { $search: searchString }, deletedAt: { $exists: false }, accountVerified: true }] }],
        })
        .exec();
    }
  }

  /**
   * Parse filter value from string to boolean
   * @param filters
   */
  private parseFilter(filters: Array<any>): Array<any> {
    return filters.map((filter) => {
      const key = Object.keys(filter)[0];
      if (filter[key] === 'True') {
        return { [key]: true };
      } else {
        return filter;
      }
    });
  }

  public async findAll(): Promise<StructureDocument[]> {
    const structures = await this.structureModel.find({ deletedAt: { $exists: false } }).exec();
    // Update structures coord and address before sending them
    await Promise.all(
      structures.map((structure: StructureDocument) => {
        // If structre has no address, add it
        if (!structure.address || structure.coord.length <= 0) {
          return this.getStructurePosition(structure).then((postition: StructureDocument) => {
            this.structureModel
              .findByIdAndUpdate(Types.ObjectId(structure._id), { address: postition.address, coord: postition.coord })
              .exec();
          });
        }
      })
    );
    return this.structureModel.find({ deletedAt: { $exists: false }, accountVerified: true }).exec();
  }

  public async populateES(): Promise<StructureDocument[]> {
    const structures = await this.structureModel.find({ deletedAt: { $exists: false } }).exec();
    await Promise.all(
      structures.map((structure: StructureDocument) => {
        this.structuresSearchService.indexStructure(structure);
      })
    );
    return structures;
  }

  public async findAllFormated(
    formationCategories: CategoriesFormations[],
    accompagnementCategories: CategoriesAccompagnement[],
    otherCategories: CategoriesOthers[]
  ): Promise<StructureDocument[]> {
    const structures = await this.structureModel.find({ deletedAt: { $exists: false } }).exec();

    // Update structures coord and address before sending them
    await Promise.all(
      structures.map((structure: StructureDocument) => {
        // If structre has no address, add it
        if (!structure.address || structure.coord.length <= 0) {
          return this.getStructurePosition(structure).then((postition: StructureDocument) => {
            this.structureModel
              .findByIdAndUpdate(Types.ObjectId(structure._id), {
                address: postition.address,
                coord: postition.coord,
              })
              .exec();
          });
        }
      })
    );
    return (
      await this.structureModel
        .find({ deletedAt: { $exists: false }, accountVerified: true })
        .select('-_id -accountVerified -otherDescription')
        .exec()
    ).map((structure) => {
      structure.proceduresAccompaniment = this.mapModules(
        structure.proceduresAccompaniment,
        accompagnementCategories.find((category) => category.id === 'proceduresAccompaniment').modules
      );
      structure.labelsQualifications = this.mapModules(
        structure.labelsQualifications,
        otherCategories.find((category) => category.id === 'labelsQualifications').modules
      );
      structure.publics = this.mapModules(
        structure.publics,
        otherCategories.find((category) => category.id === 'publics').modules
      );
      structure.accessModality = this.mapModules(
        structure.accessModality,
        otherCategories.find((category) => category.id === 'accessModality').modules
      );
      structure.publicsAccompaniment = this.mapModules(
        structure.publicsAccompaniment,
        otherCategories.find((category) => category.id === 'publicsAccompaniment').modules
      );
      structure.equipmentsAndServices = this.mapModules(
        structure.equipmentsAndServices,
        otherCategories.find((category) => category.id === 'equipmentsAndServices').modules
      );
      structure.baseSkills = this.mapFormationModules(
        structure.baseSkills,
        formationCategories.find((category) => category.id === 'baseSkills').modules
      );
      structure.accessRight = this.mapFormationModules(
        structure.accessRight,
        formationCategories.find((category) => category.id === 'accessRight').modules
      );
      structure.socialAndProfessional = this.mapFormationModules(
        structure.socialAndProfessional,
        formationCategories.find((category) => category.id === 'socialAndProfessional').modules
      );
      structure.parentingHelp = this.mapFormationModules(
        structure.parentingHelp,
        formationCategories.find((category) => category.id === 'parentingHelp').modules
      );
      structure.digitalCultureSecurity = this.mapFormationModules(
        structure.digitalCultureSecurity,
        formationCategories.find((category) => category.id === 'digitalCultureSecurity').modules
      );
      return structure;
    });
  }
  public mapFormationModules(structureModule: string[], baseModule: CategoriesFormationsModule[]): string[] {
    if (structureModule == []) {
      return [];
    }
    return structureModule.map((id) => {
      const formatedText = baseModule.find((module) => module.display_id === id)?.text;
      return formatedText ? formatedText : id;
    });
  }
  public mapModules(structureModule: string[], baseModule: CategoriesModule[]): string[] {
    if (structureModule == []) {
      return [];
    }
    return structureModule.map((id) => {
      const formatedText = baseModule.find((module) => module.id === id)?.text;
      return formatedText ? formatedText : id;
    });
  }

  public async update(idStructure: string, structure: structureDto): Promise<Structure> {
    const oldStructure = await this.findOne(idStructure);
    if (!_.isEqual(oldStructure.address, structure.address)) {
      await this.getStructurePosition(structure).then();
    }
    const result = await this.structureModel.findByIdAndUpdate(Types.ObjectId(idStructure), structure).exec();
    if (!result) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    } else {
      await this.structuresSearchService.update(structure, idStructure);
      this.sendAdminStructureNotification(
        result,
        this.mailerService.config.templates.structureModificationNotification.ejs,
        this.mailerService.config.templates.structureModificationNotification.json
      );
      this.userService.removeOutdatedStructureFromArray(idStructure);
    }
    return this.findOne(idStructure);
  }

  public async findOne(idParam: string): Promise<StructureDocument> {
    return await this.structureModel.findById(Types.ObjectId(idParam)).exec();
  }
  /**
   * Get structures positions and add marker corresponding to those positons on the map
   */
  private getStructurePosition(structure: Structure): Promise<Structure> {
    return new Promise((resolve, reject) => {
      this.getCoord(structure.address.numero, structure.address.street, structure.address.commune).subscribe(
        (res) => {
          const address = res.data.features[0];
          if (address && address.geometry) {
            structure.coord = address.geometry.coordinates;
          } else {
            Logger.error(
              `No coord found for: ${structure.address.numero} ${structure.address.street} ${structure.address.commune}`,
              'StructureService'
            );
            structure.coord = [];
          }
          resolve(structure);
        },
        (err) => {
          Logger.error(`Request error: ${err.config.url}`, 'StructureService');
          Logger.error(err);
        }
      );
    });
  }

  public async isClaimed(structureId: string, user: User): Promise<boolean> {
    const isStructureClaimed = await this.userService.isStructureClaimed(structureId);
    const isUserAlreadyClaimed = await this.userService.isUserAlreadyClaimedStructure(structureId, user.email);
    if (isStructureClaimed || isUserAlreadyClaimed) {
      return true;
    }
    return false;
  }

  /**
   * Search structure address based on data search WS
   */
  public async searchAddress(data: { searchQuery: string }): Promise<AxiosResponse<any>> {
    const req =
      'https://download.data.grandlyon.com/geocoding/photon-bal/api?q=' +
      data.searchQuery +
      '&lat=45.75&lon=4.85&lang=fr&limit=50&osm_tag=:!construction&osm_tag=:!bus_stop';
    return new Promise((resolve, reject) => {
      this.httpService
        .request({
          url: encodeURI(req),
          method: 'GET',
          headers: { 'Content-Type': 'application/json' },
        })
        .subscribe(
          (reply) => {
            return resolve(reply.data);
          },
          (err) => {
            Logger.error(`Request error: ${err.config.url}`, 'StructureService - search');
            Logger.error(err);
          }
        );
    });
  }

  /**
   * Count every value occurence of a given key
   * @param key structure key
   * @return [{id: 'key', count: 'value'}]
   */
  public async countByStructureKey(key: string, selected: { id: string; text: string }[]): Promise<any> {
    const uniqueElements = await this.structureModel.distinct(key).exec();
    return await Promise.all(
      uniqueElements.map(async (value) => {
        const keyList: FilterQuery<DocumentDefinition<StructureDocument>>[] = [];
        keyList.push({
          [key]: { $elemMatch: { $eq: value } },
          deletedAt: { $exists: false },
          accountVerified: true,
        });
        if (selected && selected.length > 0) {
          for (const val of selected) {
            keyList.push({
              [val.text]: { $elemMatch: { $eq: val.id } },
              deletedAt: { $exists: false },
              accountVerified: true,
            });
          }
        }
        return {
          id: value,
          count: await this.structureModel
            .countDocuments({
              $and: keyList,
            })
            .exec(),
        };
      })
    );
  }

  public getCoord(numero: string, address: string, zipcode: string): Observable<AxiosResponse<any>> {
    const req =
      'https://download.data.grandlyon.com/geocoding/photon-bal/api' + '?q=' + numero + ' ' + address + ' ' + zipcode;
    Logger.log(`Request : ${req}`, 'StructureService - getCoord');
    return this.httpService.get(encodeURI(req));
  }

  public async deleteOne(id: string): Promise<Structure> {
    const structure = await this.structureModel.findById(Types.ObjectId(id)).exec();
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    this.structuresSearchService.deleteIndexStructure(structure);
    structure.structureType = null;
    structure.deletedAt = DateTime.local().setZone('Europe/Paris').toString();
    this.anonymizeStructure(structure).save();
    // Remove structure from userModel
    this.userService.removeStructureIdFromUsers(structure._id);
    this.sendAdminStructureNotification(
      structure,
      this.mailerService.config.templates.structureDeletionNotification.ejs,
      this.mailerService.config.templates.structureDeletionNotification.json
    );
    return structure;
  }

  public async sendAdminStructureNotification(
    structure: StructureDocument,
    templateLocation: any,
    jsonConfigLocation: any
  ) {
    const uniqueAdminEmails = [...new Set((await this.userService.getAdmins()).map((admin) => admin.email))].map(
      (item) => {
        return { email: item };
      }
    );

    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(templateLocation);
    const jsonConfig = this.mailerService.loadJsonConfig(jsonConfigLocation);

    const html = await ejs.renderFile(ejsPath, {
      config,
      id: structure ? structure._id : 0,
      structureName: structure ? structure.structureName : '',
    });
    this.mailerService.send(uniqueAdminEmails, jsonConfig.subject, html);
  }

  private anonymizeStructure(structure: StructureDocument): StructureDocument {
    structure.contactPhone = '';
    structure.contactMail = '';
    structure.facebook = null;
    structure.twitter = null;
    structure.instagram = null;
    structure.website = '';
    return structure;
  }

  @Cron(CronExpression.EVERY_DAY_AT_4AM)
  public async checkOutdatedStructuresInfo(): Promise<void> {
    const OUTDATED_MONT_TO_CHECK = 6;
    const structureList = await this.findAll();
    const local = DateTime.local().setZone('Europe/Paris');
    // Get outdated structures
    const filteredList = structureList.filter((structure) => {
      const updateDate = DateTime.fromFormat(structure.updatedAt.split(' GMT')[0], 'EEE MMM dd yyyy HH:mm:ss');
      const diff = local.diff(updateDate, ['months']);
      if (diff.values.months > OUTDATED_MONT_TO_CHECK) {
        return true;
      }
    });
    // Get owners of outdated structures
    const ownerList = await Promise.all(
      filteredList.map(async (structure) => {
        const owner = await this.userService.isStructureClaimed(structure._id.toString());
        if (owner) {
          return { structure: structure, owner: owner };
        }
      })
    );
    // Send email if possible and update user
    ownerList
      .filter((x) => x != undefined)
      .forEach(async (data) => {
        this.userService.findOne(data.owner.email).then((user) => {
          // If mail is already sent, do not resend
          if (user.structureOutdatedMailSent.includes(data.structure._id)) {
            return;
          } else {
            this.sendOutdatedEmailToUser(data.owner.email, data.structure.structureName, data.structure._id);
            user.structureOutdatedMailSent.push(data.structure._id);
            user.save();
          }
        });
      });
  }

  /**
   * Send an email to prevent outdated
   * a new account.
   * @param user User
   */
  private async sendOutdatedEmailToUser(userEmail: string, structureName: string, id: string): Promise<any> {
    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(config.templates.structureOutdatedInfo.ejs);
    const jsonConfig = this.mailerService.loadJsonConfig(config.templates.structureOutdatedInfo.json);

    const html = await ejs.renderFile(ejsPath, {
      config,
      name: structureName,
      id: id,
    });
    this.mailerService.send(userEmail, jsonConfig.subject, html);
  }

  /**
   * Send an email to structure owner's in order to accept or decline a join request
   * @param user User
   */
  public async sendStructureJoinRequest(user: IUser, structure: StructureDocument): Promise<void> {
    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(config.templates.structureJoinRequest.ejs);
    const jsonConfig = this.mailerService.loadJsonConfig(config.templates.structureJoinRequest.json);

    const html = await ejs.renderFile(ejsPath, {
      config,
      structureName: structure.structureName,
      name: user.name,
      surname: user.surname,
      id: structure._id,
      userId: user._id,
    });
    const owners = await this.getOwners(structure._id);
    owners.forEach((owner) => {
      this.mailerService.send(owner.email, jsonConfig.subject, html);
    });
  }

  private async getOwners(structureId: string): Promise<IUser[]> {
    // Get owners of outdated structures
    return this.userService.getStructureOwners(structureId);
  }

  public async updateAccountVerified(idStructure: string): Promise<Structure> {
    const structureLinked = await this.findOne(idStructure);
    const structure = new this.structureModel(structureLinked);
    if (!structure) {
      throw new HttpException('Invalid structure', HttpStatus.NOT_FOUND);
    }
    structure.accountVerified = true;
    structure.save();
    return structure;
  }

  public async findWithOwners(
    idStructure: string,
    emailUser: string
  ): Promise<{ structure: Structure; owners: OwnerDto[] }> {
    const structure = await this.findOne(idStructure);
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    const owners = await this.userService.getStructureOwnersMails(idStructure, emailUser);
    return { structure: structure, owners: owners };
  }

  public async reportStructureError(structureId: string, content: string) {
    const structure = await this.findOne(structureId);
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    const owners = await this.userService.getStructureOwnersMails(structure._id, '');
    const admins = await this.userService.getAdmins();
    const emails = owners.map((owner) => owner.email).concat(admins.map((admin) => admin.email));
    const uniqueEmails = [...new Set(emails)];

    const emailsObject: { email: string }[] = uniqueEmails.map((item) => {
      return { email: item };
    });

    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(config.templates.structureErrorReport.ejs);
    const jsonConfig = this.mailerService.loadJsonConfig(config.templates.structureErrorReport.json);

    const html = await ejs.renderFile(ejsPath, {
      config,
      content: content,
      id: structure._id,
      structureName: structure.structureName,
    });
    this.mailerService.send(emailsObject, jsonConfig.subject, html);
  }
}
